# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.placement package."""

# FIXME: Unnecessary recursion.
from openfl.experimental.workflow.placement.placement import aggregator, collaborator
